<?php
/**
 * Created by PhpStorm.
 * User: crose
 * Date: 1/4/16
 * Time: 7:14 PM
 */

namespace qfq;

use qfq;
use qfq\Store;
use qfq\CodeException;
use qfq\UserException;
use qfq\DbException;
use qfq\Support;


require_once(__DIR__ . '/../qfq/exceptions/UserException.php');
require_once(__DIR__ . '/../qfq/exceptions/CodeException.php');
require_once(__DIR__ . '/../qfq/exceptions/DbException.php');

require_once(__DIR__ . '/../qfq/store/Store.php');
require_once(__DIR__ . '/../qfq/helper/Support.php');

/**
 * Class Database
 * @package qfq
 */
class Database {

    /**
     * @var \PDO
     */
    public $pdo = null;
    /**
     * @var Store
     */
    private $store = null;
    private $stmt = null;

    /**
     * Returns current data base handle from Store[System][SYSTEM_DBH].
     * If not exists: open database and store the new dbh in Store[System][SYSTEM_DBH]
     *
     * @throws CodeException
     * @throws UserException
     */
    public function __construct() {
        $this->store = Store::getInstance();

        $this->pdo = $this->store->getVar(SYSTEM_DBH, STORE_SYSTEM);

        if ($this->pdo === false) {
            $this->pdo = $this->dbConnect();
            $this->store->setVar(SYSTEM_DBH, $this->pdo, STORE_SYSTEM);
        }
    }

    /**
     * If not
     * @return \PDO
     * @throws UserException
     */
    private function dbConnect() {
        $dbuser = $this->store->getVar(SYSTEM_DBUSER, STORE_SYSTEM);
        $dbserver = $this->store->getVar(SYSTEM_DBSERVER, STORE_SYSTEM);
        $dbpw = $this->store->getVar(SYSTEM_DBPW, STORE_SYSTEM);
        $db = $this->store->getVar(SYSTEM_DB, STORE_SYSTEM);

        try {
            $pdo = new \PDO("mysql:host=" . $dbserver . ";dbname=" . $db, $dbuser, $dbpw, array(\PDO::ATTR_PERSISTENT => true));
        } catch (\Exception $e) {
            throw new UserException ("Error open Database 'mysql:host=" . $dbserver . ";dbname=" . $db . ";dbuser=" . $dbuser . "'': " . $e->getMessage(), ERROR_OPEN_DATABASE);
        }

        return $pdo;
    }

    /**
     * Return the number of rows returned by the last call to execute().
     *
     * If execute() has never been called, returns FALSE.
     *
     * @return mixed Number of rows returned by last call to execute(). If Database::execute()
     *     has never been called prior a call to this method, false is returned.
     */
    public function getRowCount() {
        if ($this->stmt == null) {
            return false;
        }

        return $this->stmt->rowCount();
    }

    /**
     * Get the values for a given ENUM or SET column
     *
     * @param string $table name of the table
     * @param string $columnName name of the column
     *
     * @throws UserException if the table or column does not exist, or is not of type ENUM or SET
     * @return array
     */
    public function getEnumSetValueList($table, $columnName) {

        $columnDefinition = $this->getFieldDefinitionFromTable($table, $columnName);
        $setEnumDefinition = $columnDefinition["Type"];

        // $setEnumDefinition holds now a string like
        // String:  enum('','red','blue','green')
        $len = mb_strlen($setEnumDefinition);

        # "enum('" = 6, "set('" = 5
        $tokenLength = strpos($setEnumDefinition, "'") + 1;

        // count("enum('") == 6, count("')") == 2
        $enumSetString = mb_substr($setEnumDefinition, $tokenLength, $len - (2 + $tokenLength));

        // String: ','red','blue','green

        if (($setEnumValueList = explode("','", $enumSetString)) === false) {
            return array();
        }

        return $setEnumValueList;
    }

    /**
     * Get database column definition.
     *
     * If the column is not found in the table, an exception is thrown.
     *
     * @param string $table name of the table
     *
     * @param string $columnName name of the column
     *
     * @return array the definition of the column as retrieved by Database::getTableDefinition().
     *
     * @throws UserException
     *
     */
    private function getFieldDefinitionFromTable($table, $columnName) {
        $tableDefinition = $this->getTableDefinition($table);
        foreach ($tableDefinition AS $row) {
            if ($row["Field"] == $columnName) {
                return $row;
            }
        }
        throw new DbException("Column name '$columnName' not found in table '$table'.", ERROR_COLUMN_NOT_FOUND_IN_TABLE);
    }

    /**
     * Get all column definitions for a table. Return Assoc Array:
     *
     * Field      Type                      Null    Key    Default    Extra
     * --------------------------------------------------------------------------
     * id         bigint(20)                 NO     PRI    NULL    auto_increment
     * name       varchar(128)               YES           NULL
     * firstname  varchar(128)               YES           NULL
     * gender     enum('','male','female')   NO            male
     * groups     set('','a','b','c')        NO            a
     *
     * @param string $table table to retrieve column definition from
     *
     * @return array column definition of table as returned by SHOW FIELDS FROM as associative array.
     */
    public function getTableDefinition($table) {
        return $this->sql("SHOW FIELDS FROM `$table`");
    }

    /**
     * Fires query $sql and fetches result als assoc array.
     * $mode
     *  ROW_REGULAR      return result set as assoc array. No check about number of rows.
     *  ROW_IMPLODE_ALL  return result set as one string: implode all rows and columns.  No check about number of rows.
     *  ROW_EXACT_1      return result set as assoc array. It has to be exact one row selected. If not: throw Exception,
     *  ROW_EMPTY_IS_OK  return result set as assoc array. It has to be 0 or one row selected. If not: throw Exception,
     *
     * @param $sql
     * @param string $mode
     * @param array $parameterArray
     * @param string $specificMessage
     * @return mixed|null
     * @throws \qfq\CodeException
     * @throws \qfq\DbException
     */
    public function sql($sql, $mode = ROW_REGULAR, array $parameterArray = array(), $specificMessage = '') {

        // CR often forgets to specify the $mode and use prepared statement with parameter instead.
        if (is_array($mode))
            throw new CodeException("Probably a parameter forgotten: \$mode ?");

        // for error reporting in exception
        if ($specificMessage)
            $specificMessage .= " ";

        $count = $this->prepareExecute($sql, $parameterArray);
        if ($count === false) {
            throw new DbException($specificMessage . "No idea why this error happens - please take some time and check this: $sql", ERROR_DB_GENERIC_CHECK);
        }

        if ($mode === ROW_IMPLODE_ALL) {
            return $this->fetchAll(ROW_IMPLODE_ALL);
        }

        if ($mode === ROW_REGULAR) {
            return $this->fetchAll();
        }

        if ($count < 1) {
            if ($mode === ROW_EMPTY_IS_OK) {
                return array();
            }

            throw new DbException($specificMessage . "Expected one row, got nothing: $sql", ERROR_DB_TOO_FEW_ROWS);
        }

        if ($count > 1) {
            throw new DbException($specificMessage . "Expected one row, got $count rows: $sql", ERROR_DB_TOO_MANY_ROWS);
        }

        // ROW_EXACT_1
        return $this->fetchOne();

    }

    /**
     * Execute a prepared SQL statement like SELECT, INSERT, UPDATE, DELETE, SHOW, ...
     *
     * The result of the query can be retrieved by using one of the fetch methods.
     *
     * @param string $sql SQL statement with prepared statement variable.
     * @param array $parameterArray parameter array for prepared statement execution.
     * @return mixed
     * @throws \qfq\DbException
     * @throws \qfq\UserException
     */
    public function prepareExecute($sql, array $parameterArray = array()) {

        $this->store->setVar(SYSTEM_SQL_FINAL, $sql, STORE_SYSTEM);
        $this->store->setVar(SYSTEM_SQL_PARAM_ARRAY, $parameterArray, STORE_SYSTEM);

        // Logfile
        $this->dbLog($sql, $parameterArray);

        $this->stmt = $this->pdo->prepare($sql);
        if (false === $this->stmt->execute($parameterArray)) {
            throw new DbException($this->stmt->errorInfo()[2], ERROR_DB_EXECUTE);
        }

        $count = $this->stmt->rowCount();
        $this->store->setVar(SYSTEM_SQL_COUNT, $count, STORE_SYSTEM);

        // Logfile
        $msg = (substr($sql, 0, 7) === 'INSERT ') ? 'ID: ' . $this->getLastInsertId() : 'Affected rows: ' . $count;
        $this->dbLog($msg);

        return $count;
    }

    /**
     * @param $sql
     * @param $parameterArray
     * @throws \qfq\CodeException
     */
    private function dbLog($sql, $parameterArray = array()) {
        $msg = '[' . date('r') . '][';

        if (count($parameterArray) === 0) {
            $msg .= $sql;
        } else {

            $sqlArray = explode('?', $sql);
            $ii = 0;
            foreach ($parameterArray as $value) {
                if (isset($sqlArray[$ii])) {
                    $msg .= $sqlArray[$ii++] . "'" . $value . "'";
                } else {
                    $msg = '?';
                }
            }
            if (isset($sqlArray[$ii]))
                $msg .= $sqlArray[$ii];
        }

        $msg .= ']';

        Support::log($msg);
    }

    /**
     * Returns lastInsertId
     *
     * @return string
     */
    public function getLastInsertId() {
        return $this->pdo->lastInsertId();
    }

    /**
     * Fetch all rows of the result as associative array.
     *
     * @return mixed false in case of error.
     *              Empty string is returned if the query didn't yield any rows.
     *              All rows as Multi Assoc array if $mode!=IMPLODE_ALL.
     *              All rows and all columns imploded to one string if $mode=IMPLODE_ALL
     *
     */
    public function fetchAll($mode = '') {
        if ($this->stmt == null) {
            return false;
        }

        if ($this->stmt->rowCount() === 0) {
            return ($mode === ROW_IMPLODE_ALL) ? "" : array();
        }

        $str = "";
        if ($mode === ROW_IMPLODE_ALL) {
            foreach ($this->stmt->fetchAll(\PDO::FETCH_ASSOC) as $row) {
                $str .= implode($row);
            }
            return $str;
        } else {
            return $this->stmt->fetchAll(\PDO::FETCH_ASSOC);
        }
    }

    /**
     * Fetch one row of the result as associative array.
     *
     * @return null|mixed the first row of the result of Database::execute() as associative array.
     *  If Database::execute() has never been called prior a call to this method, false is returned.
     *  Empty string is returned if the query didn't yield any rows.
     */
    public function fetchOne() {
        if ($this->stmt == null) {
            return false;
        }

        if ($this->stmt->rowCount() === 0) {
            return array();
        }

        return $this->stmt->fetch(\PDO::FETCH_ASSOC);
    }

}