<?php
/**
 * @author Rafael Ostertag <rafael.ostertag@math.uzh.ch>
 */

//namespace qfq;

//use qfq\DbException;

require_once(__DIR__ . '/AbstractDatabaseTest.php');
require_once(__DIR__ . '/../../qfq/Database.php');
require_once(__DIR__ . '/../../qfq/exceptions/DbException.php');

class DatabaseTest extends AbstractDatabaseTest {

    /**
     * @throws \qfq\DbException
     */
    public function testExecute() {

        $this->db->prepareExecute('SELECT * FROM Person');

        $this->assertEquals(2, $this->db->getRowCount());
    }

    /**
     * @throws \qfq\DbException
     */
    public function testFetchOneRowAssocEmptyResult() {
        $this->db->prepareExecute('SELECT * FROM Person WHERE id=0', []);

        $this->assertEquals(array(), $this->db->fetchOne());
    }

    /**
     * @throws \qfq\DbException
     */
    public function testFetchOne() {
        $this->db->prepareExecute('SELECT * FROM Person');

        $oneRow = $this->db->fetchOne();

        $this->assertArrayHasKey('name', $oneRow);
        $this->assertArrayHasKey('firstname', $oneRow);
        $this->assertArrayHasKey('gender', $oneRow);

        $this->assertEquals('Doe', $oneRow['name']);
        $this->assertEquals('John', $oneRow['firstname']);
    }

    /**
     * @throws \qfq\DbException
     */
    public function testFetchAll() {
        $this->db->prepareExecute('SELECT * FROM Person ORDER BY id LIMIT 2');

        $allRows = $this->db->fetchAll();

        $this->assertCount(2, $allRows);
        $this->assertEquals(1, $allRows[0]['id']);
        $this->assertEquals(2, $allRows[1]['id']);
    }

    /**
     * @throws \qfq\DbException
     */
    public function testFetchAllEmpty() {
        $this->db->prepareExecute('SELECT * FROM Person WHERE id=0 ORDER BY id');

        $this->assertEquals(array(), $this->db->fetchAll());
    }

    /**
     * @throws \qfq\DbException
     */
    public function testQuerySimple() {

        $expected = [
            [
                'id' => '1',
                'name' => 'Doe',
                'firstname' => 'John',
                'gender' => 'male',
                'groups' => 'c'
            ],
            [
                'id' => '2',
                'name' => 'Smith',
                'firstname' => 'Jane',
                'gender' => 'female',
                'groups' => 'a,c'
            ],
        ];

        // Check read rows
        $dataArray = $this->db->sql('SELECT * FROM Person ORDER BY id LIMIT 2');
        // Check count
        $this->assertEquals(2, $this->db->getRowCount());
        $this->assertEquals(2, $this->store->getVar(SYSTEM_SQL_COUNT, STORE_SYSTEM));
        // Compare rows
        $this->assertSame($expected, $dataArray);
        // Check rows

        // Same as above, but specify 'ROW_REGULAR'
        $dataArray = $this->db->sql('SELECT * FROM Person ORDER BY id LIMIT 2', ROW_REGULAR);
        $this->assertEquals(2, $this->db->getRowCount());
        // Check rows
        $this->assertSame($expected, $dataArray);

        // ROW_EXACT_1
        $dataArray = $this->db->sql('SELECT * FROM Person WHERE id=1', ROW_EXACT_1);
        $this->assertEquals(1, $this->db->getRowCount());
        // Check rows
        $this->assertSame($expected[0], $dataArray);

        // ROW_EMPTY_IS_OK - one record
        $dataArray = $this->db->sql('SELECT * FROM Person WHERE id=1', ROW_EMPTY_IS_OK);
        $this->assertEquals(1, $this->db->getRowCount());
        // Check rows
        $this->assertSame($expected[0], $dataArray);

        // ROW_EMPTY_IS_OK - no record
        $dataArray = $this->db->sql('SELECT * FROM Person WHERE id=0', ROW_EMPTY_IS_OK);
        $this->assertEquals(0, $this->db->getRowCount());
        // Check rows
        $this->assertSame(array(), $dataArray);

        // Check Implode: 2 records
        $data = $this->db->sql('SELECT * FROM Person ORDER BY id LIMIT 2', ROW_IMPLODE_ALL);
        $this->assertEquals(implode($expected[0]) . implode($expected[1]), $data);

        // Check Implode: 1 record
        $data = $this->db->sql('SELECT * FROM Person ORDER BY id LIMIT 1', ROW_IMPLODE_ALL);
        $this->assertEquals(implode($expected[0]), $data);

        // Check Implode 0 record
        $data = $this->db->sql('SELECT * FROM Person WHERE id=0 ORDER BY id LIMIT 2', ROW_IMPLODE_ALL);
        $this->assertEquals('', $data);
    }

    /**
     * @throws \qfq\DbException
     */
    public function testQuerySimpleParameter() {
        // Parameter Susbstitution by '?'
        $dataArray = $this->db->sql('SELECT * FROM Person WHERE name LIKE ? ORDER BY id', ROW_REGULAR, ['Smith']);
        // Check count
        $this->assertEquals(1, $this->db->getRowCount());


        // Parameter Susbstitution by name
        $dataArray = $this->db->sql('SELECT * FROM Person WHERE name LIKE :valueOfName ORDER BY id ', ROW_REGULAR, [':valueOfName' => 'Smith']);
        // Check count
        $this->assertEquals(1, $this->db->getRowCount());

    }

    /**
     * @expectedException \qfq\DbException
     *
     */
    public function testSanatizeException() {
        $this->db->sql('some garbage');
    }

    /**
     * @throws \qfq\DbException
     */
    public function testParameters() {
        $this->db->prepareExecute('SELECT * FROM Person WHERE id = :valueOfA', [':valueOfA' => 2]);
        $this->assertEquals(1, $this->db->getRowCount());
    }

    /**
     *
     */
    public function testGetSetValueList() {
        $valueList = $this->db->getEnumSetValueList('Person', 'gender');
        $expected = [0 => '', 1 => 'male', 2 => 'female'];
        $this->assertEquals($expected, $valueList);

        $expected = [0 => '', 1 => 'a', 2 => 'b', 3 => 'c'];
        $valueList = $this->db->getEnumSetValueList('Person', 'groups');
        $this->assertEquals($expected, $valueList);
    }

    /**
     * @throws \qfq\DbException
     */
    public function testGetLastInsertId() {

        $sql = "INSERT INTO Person (id, name, firstname, gender, groups) VALUES (NULL, 'Doe', 'Jonni', 'male','')";

        $this->db->prepareExecute($sql);
        $this->assertEquals(3, $this->db->getLastInsertId());
    }

    /**
     *
     */
    public function testGetTableDefinition() {
        $expected = [
            ['Field' => 'id', 'Type' => 'bigint(20)', 'Null' => 'NO', 'Key' => 'PRI', 'Default' => '', 'Extra' => 'auto_increment'],
            ['Field' => 'name', 'Type' => 'varchar(128)', 'Null' => 'YES', 'Key' => '', 'Default' => '', 'Extra' => ''],
            ['Field' => 'firstname', 'Type' => 'varchar(128)', 'Null' => 'YES', 'Key' => '', 'Default' => '', 'Extra' => ''],
            ['Field' => 'gender', 'Type' => "enum('','male','female')", 'Null' => 'NO', 'Key' => '', 'Default' => 'male', 'Extra' => ''],
            ['Field' => 'groups', 'Type' => "set('','a','b','c')", 'Null' => 'NO', 'Key' => '', 'Default' => '', 'Extra' => ''],
        ];

        $this->assertEquals($expected, $this->db->getTableDefinition('Person'));
    }


    /**
     * @throws Exception
     */
    protected function setUp() {
        parent::setUp();

        $this->executeSQLFile(__DIR__ . '/fixtures/Generic.sql', true);
    }
}
